#include <Communication/BTSerialCommunicationManager.h>

// Adapted from Finally Functional's SerialBT implementation

BTSerialCommunicationManager::BTSerialCommunicationManager(const VRBTSerialConfiguration_t& configuration, std::unique_ptr<IEncodingManager> encodingManager)
    : m_btSerialConfiguration(configuration), m_encodingManager(std::move(encodingManager)), m_isConnected(false){};

void BTSerialCommunicationManager::Connect() {
  DriverLog("Trying to connect to bluetooth");

  // We're not yet connected
  m_isConnected = false;

  // Try to connect
  if (!getPairedDeviceBtAddress())  // find an device paired with this machine
  {
    DriverLog("Error getting Bluetooth address");
    return;
  }
  if (!startupWindowsSocket())  // initialize windows sockets
  {
    DriverLog("Error Initializing windows sockets");
    return;
  }
  if (!connectToDevice())  // initialize BT windows socket for connecting to device
  {
    DriverLog("Error connecting to Bluetooth device");
    return;
  } else {
    // If everything went fine we're connected
    m_isConnected = true;
    DriverLog("Connected to bluetooth!");
  }
}

void BTSerialCommunicationManager::BeginListener(const std::function<void(VRCommData_t)>& callback) {
  DriverLog("Begun listener");
  m_threadActive = true;
  m_serialThread = std::thread(&BTSerialCommunicationManager::ListenerThread, this, callback);
}

void BTSerialCommunicationManager::ListenerThread(const std::function<void(VRCommData_t)>& callback) {
  // DebugDriverLog("In listener thread");
  std::this_thread::sleep_for(std::chrono::milliseconds(ARDUINO_WAIT_TIME));

  while (m_threadActive) {
    std::string receivedString;
    bool readSuccessful = ReceiveNextPacket(receivedString);
    if (readSuccessful) {
      try {
        VRCommData_t commData = m_encodingManager->Decode(receivedString);
        callback(commData);
        sendMessageToDevice();
      } catch (const std::invalid_argument& ia) {
        DriverLog("Received error from encoding manager. Skipping...");
      }
    } else {
      DriverLog("Detected that arduino has disconnected! Stopping listener...");
      // We should probably do more logic for trying to reconnect to the arduino
      // For now, it should be obvious to people that the arduinos have disconnected
      m_threadActive = false;
    }
  }
}

bool BTSerialCommunicationManager::ReceiveNextPacket(std::string& buff) {
  char nextChar[1];
  do {
    int recievedMessageLength = 1;
    int recieveResult = recv(m_btClientSocket, nextChar, recievedMessageLength, 0);  // if your socket is blocking, this will block until a
    if (recieveResult < 0)                                                           // a message is recieved. If not, it will return right
    {                                                                                // away
      continue;
    }
    buff += nextChar[0];
  } while (nextChar[0] != '\n' || buff.length() < 1);

  //("Packet Recieved! Length: %i, %s", buff.length(), buff.c_str());

  return true;
}

void BTSerialCommunicationManager::QueueSend(const VRFFBData_t& data) {
    std::lock_guard<std::mutex> lock(m_writeMutex);

    m_writeString = m_encodingManager->Encode(data);
}

void BTSerialCommunicationManager::Disconnect() {
  if (m_isConnected) {
    if (m_threadActive) {
      m_threadActive = false;
      m_serialThread.join();
    }
    m_isConnected = false;
    // CloseHandle(m_hSerial);

    // Disconnect
    if (shutdown(m_btClientSocket, 2) == SOCKET_ERROR) {
      DriverLog("Could not disconnect socket from bluetooth device. Error %ld", WSAGetLastError());
    } else
      DriverLog("Disconnected from socket successfully.");
  }
}
// May want to get a heartbeat here instead?
bool BTSerialCommunicationManager::IsConnected() { return m_isConnected; }

/// <summary>
/// Gets the bluetooth devices paired with this machine and
/// finds a device. If it finds one, its BT address is stored.
/// </summary>
bool BTSerialCommunicationManager::getPairedDeviceBtAddress() {
  BLUETOOTH_DEVICE_SEARCH_PARAMS btDeviceSearchParameters = {
      sizeof(BLUETOOTH_DEVICE_SEARCH_PARAMS),  // size of object
      1,                                       // return authenticated devices
      0,                                       // return remembered devices
      1,                                       // return unknown devices
      1,                                       // return connected devices
      1,                                       // issue inquery
      2,                                       // timeout multipler. Multiply this value by 1.28 seconds to get timeout.
      NULL                                     // radio handler
  };
  BLUETOOTH_DEVICE_INFO btDeviceInfo = {sizeof(BLUETOOTH_DEVICE_INFO), 0};  // default
  HBLUETOOTH_DEVICE_FIND btDevice = NULL;
  btDevice = BluetoothFindFirstDevice(&btDeviceSearchParameters, &btDeviceInfo);  // returns first BT device connected to this machine
  if (btDevice == NULL) {
    DriverLog("Could not find any bluetooth devices.");
    return false;
  }
  do {
    std::wstring thiswstring = std::wstring(m_btSerialConfiguration.name.begin(), m_btSerialConfiguration.name.end());

    m_wcDeviceName = (WCHAR*)(thiswstring.c_str());
    if (wcscmp(btDeviceInfo.szName, m_wcDeviceName) == 0) {  //
      DriverLog("Bluetooth Device found!\r\n");
      if (btDeviceInfo.fAuthenticated)  // I found that if fAuthenticated is true it means the device is paired.
      {
        DriverLog("Bluetooth Device is authenticated.\r\n");
        m_deviceBtAddress = btDeviceInfo.Address.ullLong;
        return true;
      } else {
        DriverLog("This Bluetooth Device is not authenticated. Please pair with it first.\r\n");
      }
    }
  } while (BluetoothFindNextDevice(btDevice, &btDeviceInfo));  // loop through remaining BT devices connected to this machine

  DriverLog("Could not find a paired Bluetooth Device with name %s", m_btSerialConfiguration.name.c_str());
  return false;
}

bool BTSerialCommunicationManager::startupWindowsSocket() {
  WORD wVersionRequested;
  WSADATA wsaData;
  wVersionRequested = MAKEWORD(2, 2);
  int wsaStartupError = WSAStartup(wVersionRequested, &wsaData);  // call this before using BT windows socket.
  if (wsaStartupError != 0) {
    DebugDriverLog("WSAStartup failed with error: %d", wsaStartupError);
    return false;
  }
  return true;
}

bool BTSerialCommunicationManager::connectToDevice() {
  m_btClientSocket = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM);  // initialize BT windows socket
  memset(&m_btSocketAddress, 0, sizeof(m_btSocketAddress));
  m_btSocketAddress.addressFamily = AF_BTH;
  m_btSocketAddress.serviceClassId = RFCOMM_PROTOCOL_UUID;
  m_btSocketAddress.port = 0;                                                                    // port needs to be 0 if the remote device is a client. See references.
  m_btSocketAddress.btAddr = m_deviceBtAddress;                                                   // this is the BT address of the remote device.
  if (connect(m_btClientSocket, (SOCKADDR*)&m_btSocketAddress, sizeof(m_btSocketAddress)) != 0)  // connect to the BT device.
  {
    DriverLog("Could not connect socket to Bluetooth Device. Error %ld", WSAGetLastError());
    return false;
  }
  unsigned long nonBlockingMode = 1;
  if (ioctlsocket(m_btClientSocket, FIONBIO, (unsigned long*)&nonBlockingMode) != 0)  // set the socket to be non-blocking, meaning
  {                                                                                   // it will return right away when sending/recieving
    DriverLog("Could not set socket to be non-blocking.");
    return false;
  }
  return true;
}

bool BTSerialCommunicationManager::sendMessageToDevice() {
  std::lock_guard<std::mutex> lock(m_writeMutex);
  const char* message = m_writeString.c_str();
  //DebugDriverLog("Sending %s to Bluetooth Device.", m_writeString.c_str());
  int sendResult = send(m_btClientSocket, message, (int)strlen(message), 0);  // send your message to the BT device
  if (sendResult == SOCKET_ERROR) {
    DriverLog("Sending to Bluetooth Device failed. Error code %d", WSAGetLastError());
    closesocket(m_btClientSocket);
    WSACleanup();
    return false;
  }
  return true;
}